/* dirsel.c
 * Copyright (C) 2002-2003 Pascal Eberhard <pascal.ebo@netcourrier.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>      /* define FILENAME_MAX */
#include <dirent.h>     /* scandir */
#include <string.h>     /* strcmp */
#include "common.h"
#include "gdk/gdkkeysyms.h"
#include "errno.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

struct apwalapp_t;

typedef struct dirsel_t
{
  struct apwalapp_t     *apwal;
  gchar                  dirname[2048];
  gchar                  olddir[2048];
  void                  *user_data;
  void (*compute_result) (gchar *olddir, gchar *newdir, void *data);
  
  GtkWidget    *window;
  GtkWidget      *vbox;
  GtkWidget        *edit_hbox;
  GtkWidget          *edit_label;
  GtkWidget          *edit_entry;
  GtkWidget        *dir_scroll;
  GtkWidget          *dir_tree;
  GtkWidget        *btn_bbox;
  GtkWidget          *btn_cancel;
  GtkWidget          *btn_ok;
} dirsel_t;


static GtkWidget *dirsel_dirtree_new(dirsel_t *dirsel);
static int dirsel_scandir_select_directory_only(const struct dirent *de);
static void dirsel_dirtree_scan(dirsel_t *dirsel);
static const char* dirsel_dirtree_suggest_expand(dirsel_t *dirsel,
                                                 const char *current);
static void dirsel_dirtree_refresh(dirsel_t *dirsel);
static void dirsel_build_interface(dirsel_t *dirsel);

static int str_last_slash_pos(const char *str);

static void dirsel_clicked_cancel(GtkWidget *widget, dirsel_t *dirsel);
static void dirsel_clicked_ok(GtkWidget *widget, dirsel_t *dirsel);
static void dirsel_clicked_close(GtkWidget *widget, dirsel_t *dirsel);

// ----------------------------------------------------------------------------
int str_last_slash_pos(const char *str)
{
  int i;
  int last_slash_pos;
  int str_len;

  str_len = strlen(str);
  last_slash_pos = -1;
  for (i = str_len - 1; i >= 0; i--)
  {
    if (str[i] == '/')
    {
      last_slash_pos = i;
      break;
    }
  }
  return last_slash_pos;
}
// ----------------------------------------------------------------------------
// try to find the longer valid directory in the entry
static const char* first_valid_dir(const char *dir)
{
  static char valid_dir[2048];
  int         valid_len;
  struct stat st;
  int         last_slash_pos;
  int         cc;

  g_assert(dir != NULL);
  TRACE("dir:%s", dir);

  if (dir != valid_dir)         // it is possible to recall the function
    strcpy(valid_dir, dir);     // with the result of the last call...
  
  cc = stat(valid_dir, &st);  //FIXME: link to directory
  while (cc == -1 || (!S_ISDIR(st.st_mode)))
  {
    last_slash_pos = str_last_slash_pos(valid_dir);
    if (last_slash_pos == -1)
      return NULL;
    if (last_slash_pos == 0)
    {
      valid_dir[1] = '\0';
      return valid_dir;
    }
    valid_dir[last_slash_pos] = '\0';
    cc = stat(valid_dir, &st); // FIXME: link to directory
  }
  valid_len = strlen(valid_dir);
  if (valid_dir[valid_len-1] != '/')
  {
    valid_dir[valid_len] = '/';
    valid_dir[valid_len+1] = '\0';
  }
  return valid_dir;
}
// ----------------------------------------------------------------------------
dirsel_t *dirsel_new(void)
{
  dirsel_t *dirsel;
  dirsel = (dirsel_t *) g_malloc(sizeof(dirsel_t));
  g_assert(dirsel != NULL);
  dirsel->dirname[0] = '\0';
  dirsel_build_interface(dirsel);
  //gtk_grab_remove(filesel->filesel); /*modal*/
  return dirsel;
}

// ----------------------------------------------------------------------------
gchar* dirsel_select(dirsel_t *dirsel, gchar *olddir,
             void (*compute_result) (gchar *olddir, gchar *newdir, void *data),
             void *data)
{
  gchar *newdir = NULL;
  g_assert(dirsel != NULL && olddir!= NULL && compute_result != NULL);
  
  //FIXME: if (initial_dir == NULL)
  dirsel->compute_result = compute_result;
  strcpy(dirsel->olddir, olddir);
  dirsel->user_data = data;
  gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), olddir);
  gtk_editable_set_position(GTK_EDITABLE(dirsel->edit_entry), -1);

  dirsel_dirtree_refresh(dirsel);
  gtk_grab_add(dirsel->window); 
  gtk_widget_show(dirsel->window);
  return newdir;
}

// ----------------------------------------------------------------------------
static int dirsel_scandir_select_directory_only(const struct dirent *de)
{
  g_assert(de != NULL);
  if (!strcmp(".", de->d_name))
    return 0;
  if ((de->d_name[0] == '.') && strcmp("..", de->d_name)!=0)
    return 0;           // don't show hidden directory but show .. anyway
  if (de->d_type != DT_DIR)
    return 0;
  // FIXME: link to directory
  return 1;
}
// ----------------------------------------------------------------------------
static void dirsel_dirtree_scan(dirsel_t *dirsel)
{
  int             n;
  int             i;
  struct dirent **dirlist;
  GtkTreeIter     iter;
  GtkTreeModel   *model;
  const gchar    *entry;
  const char     *dir;
  
  entry = gtk_entry_get_text(GTK_ENTRY(dirsel->edit_entry));
  TRACE("entry:%s", entry);
  model = gtk_tree_view_get_model(GTK_TREE_VIEW(dirsel->dir_tree));
  gtk_tree_store_clear(GTK_TREE_STORE(model));
  
  dir = first_valid_dir(entry);

  if (dir == NULL)
  {
    gtk_tree_store_append(GTK_TREE_STORE(model), &iter, NULL);
    gtk_tree_store_set(GTK_TREE_STORE(model), &iter, 0, "/", -1);
  }
  else
  {
    n = scandir(dir, &dirlist, dirsel_scandir_select_directory_only, alphasort);
    if (n < 0)
    {
      WARN("scandir:%d, errno:%d, dir:%s", n, errno, dir);
    }
    else
    {
      i = 0;
      while (i < n)
      {
        gtk_tree_store_append(GTK_TREE_STORE(model), &iter, NULL);
        gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
                           0, dirlist[i]->d_name, -1);
        free(dirlist[i]);
        i++;
      }
      free(dirlist);
    }
  }
}
// ----------------------------------------------------------------------------
static const char* dirsel_dirtree_suggest_expand(dirsel_t *dirsel, const char *current)
{
  GtkTreeModel     *model;
  //GtkTreeSelection *selection;
  GtkTreeIter       iter;
  //GtkTreeIter       iter_parent;
  gboolean    ok;
  gchar      *dir;
  gboolean    match = FALSE;
  int         dir_len, cur_len;
  static char suggest[128];     // fix an abitrary limit. if bigger return NULL 

  model = gtk_tree_view_get_model(GTK_TREE_VIEW(dirsel->dir_tree));
  ok = gtk_tree_model_iter_children(model, &iter, NULL);
  while (ok && !match)
  {
    gtk_tree_model_get(model, &iter, 0, &dir, -1);
    if (strstr(dir, current) == dir)
      match = TRUE;
    else
      g_free(dir);
    ok = gtk_tree_model_iter_next(model, &iter);
  }
  if (!match)
    return NULL;
  
  dir_len = strlen(dir);
  cur_len = strlen(current);
  if (dir_len == cur_len)       // it's a match ! suggest that it is ok !
    suggest[0] = '\0';
  else                          // propose the expand string
  {
    if (dir_len - cur_len >= 128)
    {
      WARN("suggestion bigger that 128 caracters... strange!"
           "dir:%s len:%d, current:%s len:%d", dir, dir_len, current, cur_len);
      return NULL;
    }
    strcpy(suggest, &dir[cur_len]);
  }
  TRACE("%s, %s, suggest:%s", current, dir, suggest);
  g_free(dir);
  return suggest;
}
// ----------------------------------------------------------------------------
static void dirsel_dirtree_refresh(dirsel_t *dirsel)
{
  dirsel_dirtree_scan(dirsel);
}

// ----------------------------------------------------------------------------
static void dirsel_edit_entry_complete(dirsel_t *dirsel)
{
  const gchar *str;
  int          str_len;
  int          cc;
  struct stat  st;
  int          last_slash_pos;
  const gchar *suggest;
  int          insert_pos;
  gchar       *insert_text;
  int          insert_text_len;

  TRACE("%s", "");
  str = gtk_entry_get_text(GTK_ENTRY(dirsel->edit_entry));
  str_len = strlen(str);
  last_slash_pos = str_last_slash_pos(str);

  if (last_slash_pos == -1)
    return;
  if (last_slash_pos == str_len - 1)
    return;

  cc = stat(str, &st);  //FIXME: link to directory
  if ((cc != -1) && (S_ISDIR(st.st_mode)))
    suggest = "";
  else
  {
    suggest = dirsel_dirtree_suggest_expand(dirsel, &str[last_slash_pos+1]);
    if (suggest == NULL)
      return;
  }

  insert_text_len = strlen(suggest) + 1 /*'/'*/;
  insert_text = (char *) g_malloc(insert_text_len + 1 /*'\0'*/);
  strcpy(insert_text, suggest);
  strcat(insert_text, "/");
  
  //g_signal_handlers_block_by_func(GTK_OBJECT(dirsel->edit_entry),
  //                                dirsel_edit_entry_insert_text, dirsel);
  insert_pos = str_len;
  gtk_editable_insert_text(GTK_EDITABLE(dirsel->edit_entry),
                           insert_text, insert_text_len, &insert_pos);
  gtk_editable_set_position(GTK_EDITABLE(dirsel->edit_entry), insert_pos);
  //g_signal_handlers_unblock_by_func(GTK_OBJECT(dirsel->edit_entry),
  //                                   dirsel_edit_entry_insert_text, dirsel);
  
  TRACE("end %s", "");
  return;
}

// ----------------------------------------------------------------------------
static void dirsel_dirtree_row_activated(GtkTreeView *treeview,
                                         GtkTreePath *path,
                                         GtkTreeViewColumn *column,
                                         dirsel_t *dirsel)
{
  GtkTreeModel      *model;
  GtkTreeIter        iter;
  const gchar       *entry;
  const char        *entry_valid_dir;
  char              *selected_dir;
  char              *newdir;
  int                newdir_len;
  int                last_slash_pos;

  TRACE("path:%s", gtk_tree_path_to_string(path));
  model = gtk_tree_view_get_model(GTK_TREE_VIEW(dirsel->dir_tree)); 
  gtk_tree_model_get_iter(model, &iter, path);
  gtk_tree_model_get(model, &iter, 0, &selected_dir, -1);
  
  entry = gtk_entry_get_text(GTK_ENTRY(dirsel->edit_entry));
  entry_valid_dir = first_valid_dir(entry);
  if (entry_valid_dir == NULL)      /* only the root could be selected here */
  {
    gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), "/");
    g_free(selected_dir);
  }
  else
  {
    if (strcmp(selected_dir, "..") == 0)
    {   // take the parent of the current directory
      newdir_len = strlen(entry_valid_dir);
      newdir = g_strdup(entry_valid_dir);
      newdir[newdir_len-1]='\0';
      last_slash_pos = str_last_slash_pos(newdir);
      if (last_slash_pos == -1)
      {
        gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), "/");
        g_free(selected_dir);
        g_free(newdir);
      }
      else
      {
        newdir[last_slash_pos+1] = '\0';
        gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), newdir);
        g_free(selected_dir);
        g_free(newdir);
      }
    }
    else
    {
      newdir_len = strlen(entry_valid_dir) + strlen(selected_dir)+ 2/*'/\0'*/;
      newdir = (char*) g_malloc(newdir_len);
      strcpy(newdir, entry_valid_dir);
      strcat(newdir, selected_dir);
      strcat(newdir, "/");
      gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), newdir);
      g_free(selected_dir);
      g_free(newdir);
    }
  }
}
// ----------------------------------------------------------------------------
static void dirsel_dirtree_selection_changed(GtkWidget *w,
                                             dirsel_t *dirsel)
{
  //TRACE("%s", "");
}
// ----------------------------------------------------------------------------
static void dirsel_edit_entry_changed(GtkWidget *w, dirsel_t *dirsel)
{
  TRACE("%s", "");
  dirsel_dirtree_refresh(dirsel);
  return;
}

// ----------------------------------------------------------------------------
static gboolean dirsel_entry_key_press(GtkWidget *w,
                                       GdkEventKey *event, dirsel_t *dirsel)
{
  if ((event->keyval == GDK_Tab || event->keyval == GDK_KP_Tab) &&
      (event->state & gtk_accelerator_get_default_mod_mask ()) == 0)
  {
    dirsel_edit_entry_complete(dirsel);
    return TRUE;
  }
  return FALSE;
}

// ----------------------------------------------------------------------------
static gboolean dirsel_treeview_key_press(GtkWidget *w,
                                          GdkEventKey *event, dirsel_t *dirsel)
{
  const gchar       *entry;
  const char        *entry_valid_dir;
  gchar             *newdir;
  int                len;
  int                last_slash_pos;
  if ((event->keyval == GDK_BackSpace) &&
      (event->state & gtk_accelerator_get_default_mod_mask ()) == 0)
  {
    TRACE("%s", "");
    entry = gtk_entry_get_text(GTK_ENTRY(dirsel->edit_entry));
    entry_valid_dir = first_valid_dir(entry);
    if (entry_valid_dir == NULL)
      return TRUE;
    else
    {   // take the parent of the current directory
      len = strlen(entry_valid_dir);
      newdir = g_strdup(entry_valid_dir);
      newdir[len-1]='\0';
      last_slash_pos = str_last_slash_pos(newdir);
      if (last_slash_pos == -1)
      {
        g_free(newdir);
        return TRUE;
      }
      newdir[last_slash_pos+1] = '\0';
      gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), newdir);
      g_free(newdir);
    }
    return TRUE;
  }
  return FALSE;
}

// ----------------------------------------------------------------------------
static void dirsel_clicked_cancel(GtkWidget *widget, dirsel_t *dirsel)
{
  gtk_widget_hide(dirsel->window);
  gtk_grab_remove(dirsel->window);
  dirsel->compute_result(dirsel->olddir, NULL, dirsel->user_data);
}
// ----------------------------------------------------------------------------
static void dirsel_clicked_ok(GtkWidget *widget, dirsel_t *dirsel)
{
  const gchar *entry;
  gtk_widget_hide(dirsel->window);
  gtk_grab_remove(dirsel->window);
  entry = gtk_entry_get_text(GTK_ENTRY(dirsel->edit_entry));
  strcpy(dirsel->dirname, entry);
  dirsel->compute_result(dirsel->olddir, dirsel->dirname, dirsel->user_data);
}
// ----------------------------------------------------------------------------
static void dirsel_clicked_close(GtkWidget *widget, dirsel_t *dirsel)
{
  gtk_widget_hide(dirsel->window);
  gtk_grab_remove(dirsel->window);
  dirsel->compute_result(dirsel->olddir, NULL, dirsel->user_data);
}
#if 0
// ----------------------------------------------------------------------------
static void dirsel_edit_entry_insert_text(GtkWidget *w,
                                          const gchar *text, gint len,
                                          gint *pos, dirsel_t *dirsel)
{ TRACE("%s", "");
}
// ----------------------------------------------------------------------------
static void dirsel_edit_entry_delete_text(GtkWidget *w,
                                          gint start, gint end,
                                          dirsel_t *dirsel)
{ TRACE("%s", "");
}
#endif

// ----------------------------------------------------------------------------
static GtkWidget *dirsel_dirtree_new(dirsel_t *dirsel)
{
  GtkTreeStore      *store;
  GtkWidget         *treeview;
  GtkTreeSelection  *selection;
  GtkCellRenderer   *renderer;
  GtkTreeViewColumn *column;

  store = gtk_tree_store_new(1, G_TYPE_STRING);
  treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
  //g_object_unref(G_OBJECT(store));
  gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
  gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(treeview), FALSE);
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
  g_signal_connect(G_OBJECT(selection), "changed",
                   G_CALLBACK(dirsel_dirtree_selection_changed), dirsel);
  g_signal_connect(G_OBJECT(treeview), "row_activated",
                   G_CALLBACK (dirsel_dirtree_row_activated), dirsel);
  g_signal_connect(G_OBJECT(treeview), "key_press_event",
                   G_CALLBACK(dirsel_treeview_key_press), dirsel);
  // directory cell
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new_with_attributes( "Directory", renderer,
                          "text", 0, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
  gtk_tree_view_set_expander_column(GTK_TREE_VIEW(treeview), column);
  return treeview;
}

// ----------------------------------------------------------------------------
static void dirsel_build_interface(dirsel_t *dirsel)
{
  g_assert(dirsel != NULL);
  
  // window
  dirsel->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(dirsel->window),"Apwal - Directory Selector");
  gtk_window_resize(GTK_WINDOW(dirsel->window), 400, 400);
  
  g_signal_connect(G_OBJECT(dirsel->window), "delete_event",
                   G_CALLBACK(dirsel_clicked_close), dirsel);
  // vbox
  dirsel->vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(dirsel->vbox), 4);
  gtk_container_add(GTK_CONTAINER(dirsel->window), dirsel->vbox);
  gtk_widget_show(dirsel->vbox);

  // edit_hbox
  dirsel->edit_hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(dirsel->edit_hbox), 4);
  gtk_box_pack_start(GTK_BOX(dirsel->vbox), dirsel->edit_hbox, FALSE, FALSE, 0);
  gtk_widget_show(dirsel->edit_hbox);
  // edit_label
  //dirsel->edit_label = gtk_label_new("Directory: ");
  //gtk_box_pack_start(GTK_BOX(dirsel->edit_hbox), dirsel->edit_label,
  //                   FALSE, FALSE, 0);
  //gtk_widget_show(dirsel->edit_label);
  // edit_entry
  dirsel->edit_entry = gtk_entry_new();
  g_signal_connect(G_OBJECT(dirsel->edit_entry), "changed",
                   G_CALLBACK(dirsel_edit_entry_changed), dirsel);
  //g_signal_connect(G_OBJECT(dirsel->edit_entry), "insert-text",
  //                 G_CALLBACK(dirsel_edit_entry_insert_text), dirsel);
  //g_signal_connect(G_OBJECT(dirsel->edit_entry), "delete-text",
  //                 G_CALLBACK(dirsel_edit_entry_delete_text), dirsel);
  g_signal_connect(G_OBJECT(dirsel->edit_entry), "key_press_event",
                   G_CALLBACK(dirsel_entry_key_press), dirsel);


  gtk_entry_set_text(GTK_ENTRY(dirsel->edit_entry), "");
  gtk_box_pack_start(GTK_BOX(dirsel->edit_hbox), dirsel->edit_entry,
                     TRUE, TRUE, 0);
  gtk_widget_show(dirsel->edit_entry);

  // dir_scroll
  dirsel->dir_scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(dirsel->dir_scroll),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_set_border_width(GTK_CONTAINER(dirsel->dir_scroll), 4);
  gtk_box_pack_start(GTK_BOX(dirsel->vbox), dirsel->dir_scroll, TRUE, TRUE, 0);
  gtk_widget_show(dirsel->dir_scroll);

  // dir_tree
  dirsel->dir_tree = dirsel_dirtree_new(dirsel);
  
  gtk_container_add(GTK_CONTAINER(dirsel->dir_scroll), dirsel->dir_tree);
  gtk_widget_show(dirsel->dir_tree);


  // btn_bbox
  dirsel->btn_bbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(dirsel->btn_bbox),GTK_BUTTONBOX_END);
  gtk_box_set_spacing(GTK_BOX(dirsel->btn_bbox), 10);
  gtk_container_set_border_width(GTK_CONTAINER(dirsel->btn_bbox), 4);
  gtk_box_pack_end(GTK_BOX(dirsel->vbox), dirsel->btn_bbox, FALSE, FALSE, 0);
  gtk_widget_show(dirsel->btn_bbox);
  // btn_cancel
  dirsel->btn_cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  GTK_WIDGET_SET_FLAGS(dirsel->btn_cancel, GTK_CAN_DEFAULT);
  g_signal_connect(G_OBJECT(dirsel->btn_cancel), "clicked",
                   G_CALLBACK(dirsel_clicked_cancel), dirsel);
  gtk_box_pack_end(GTK_BOX(dirsel->btn_bbox), dirsel->btn_cancel, TRUE, TRUE,0);
  GTK_WIDGET_SET_FLAGS(dirsel->btn_cancel, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(dirsel->btn_cancel);
  gtk_widget_show(dirsel->btn_cancel);
  // btn_ok
  dirsel->btn_ok = gtk_button_new_from_stock(GTK_STOCK_OK);
  GTK_WIDGET_SET_FLAGS(dirsel->btn_ok, GTK_CAN_DEFAULT);
  g_signal_connect(G_OBJECT(dirsel->btn_ok), "clicked",
                   G_CALLBACK(dirsel_clicked_ok), dirsel);
  gtk_box_pack_end(GTK_BOX(dirsel->btn_bbox), dirsel->btn_ok, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(dirsel->btn_ok, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(dirsel->btn_ok);
  gtk_widget_show(dirsel->btn_ok);
  
}
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
